function [mu, Sigma, Sigma_inv, acc, preds] = arow(train_data, train_labels, test_data, test_labels, r, iters)

% function [mu, Sigma, Sigma_inv, acc, preds] = 
%          arow(train_data, train_labels, test_data, test_labels, r, iters)
%
% Train the AROW classification model.
%
% Inputs: train_data   - (N x D) training matrix, where N denotes number of 
%                        data points and D is their dimensionality
%         train_labels - label vector of length N, each element having
%                        value +1/-1 and corresponding to one data point
%         test_data    - (N_test x D) testing matrix, can be empty matrix
%         test_labels  - test label vector of length N_test, can be empty
%         r            - regularization parameter (defaults to 0.1)
%         iters        - number of training epochs (defaults to 1)
%
% Output: mu           - decision model, D-dimensional vector
%         Sigma        - (D x D) covariance matrix of the decision model
%         Sigma_inv    - (D x D) inverse matrix of the cov. matrix
%         acc          - accuracy (if test set provided)
%         preds        - prediction vector of length N_test (if test set
%                        provided)
%
% Author: Nemanja Djuric (nemanja@temple.edu), according to the paper
%         Crammer K. et al., 'Adaptive Regularization of Weight Vectors', 
%         NIPS, vol. 22, pp. 414422, 2009.
% Date  : February 20th, 2016

if (nargin < 3)
    test_data = [];
    test_labels = [];
    r = 0.1;
    iters = 1;
end;
if (nargin < 5)
    r = 0.1;
    iters = 1;
end;
if (nargin < 6)
    iters = 1;
end;

[train_num_points, train_num_dim] = size(train_data);
if (~isempty(test_data))    
    [test_num_points, test_num_dim] = size(test_data);
    if (train_num_dim ~= test_num_dim)
        error('Train and test data should be of the same dimensionality.');
    end;
end;

% Initialize the parameters.
mu = zeros(train_num_dim, 1);
Sigma = eye(train_num_dim);

for iter = 1 : iters
    % Shuffle the data each iteration.
    [~, index] = sort(rand(train_num_points, 1));
    train_data = train_data(index, :);
    train_labels = train_labels(index);
    
    for i = 1 : train_num_points
        curr_point = train_data(i, :)';
        
        pred = mu' * curr_point;
        if (train_labels(i) * pred < 1)
            beta = 1 / (curr_point' * Sigma * curr_point + r);
            alpha = beta * (1 - train_labels(i) * pred);
            
            mu = mu + alpha * train_labels(i) * Sigma * curr_point;
            Sigma = Sigma - beta * (Sigma * curr_point) * (curr_point' * Sigma);
        end;
    end;
end;
% Each mapper independently computes its inverse, much faster than to do it
% on the reducer.
Sigma_inv = inv(Sigma);

% Finished training, test the model.
if (~isempty(test_data))
    num_mistakes = 0;
    preds = zeros(test_num_points, 1);
    for i = 1 : test_num_points
        curr_point = test_data(i, :)';

        preds(i) = mu' * curr_point;
        if (test_labels(i) * preds(i) <= 0)
            num_mistakes = num_mistakes + 1;
        end;
    end;
    acc = 100 * (1 - num_mistakes / test_num_points);
end;