function [mu_star, Sigma_star, accuracy] = arow_mapreduce(num_mappers, train_data, train_labels, test_data, test_labels, r, iters)

% function [mu_star, Sigma_star, accuracy] = arow_mapreduce(
%  num_mappers, train_data, train_labels, test_data, test_labels, r, iters)
%
% Train the AROW-MapReduce classification model.
%
% Inputs: num_mappers  - number of mappers used to train the model
%         train_data   - (N x D) training matrix, where N denotes number of 
%                        data points and D is their dimensionality
%         train_labels - label vector of length N, each element having
%                        value +1/-1 and corresponding to one data point
%         test_data    - (N_test x D) testing matrix, can be empty matrix
%         test_labels  - test label vector of length N_test, can be empty
%         r            - regularization parameter (defaults to 0.1)
%         iters        - number of training epochs (defaults to 1)
%
% Output: mu_star      - decision model after aggregation on the reducer, 
%                        D-dimensional vector
%         Sigma_star   - (D x D) covariance matrix of the decision model
%         accuracy     - classification accuracy
%
% Author: Nemanja Djuric (nemanja@temple.edu), according to the paper
%         Djuric N. et al., 'Distributed Confidence-Weighted Classification 
%         on MapReduce', IEEE BigData, Santa Clara, CA, USA, 2013.
% Date  : February 20th, 2016

if (nargin < 6)
    r = 0.1;
    iters = 1;
end;
if (nargin < 7)
    iters = 1;
end;

%% Send a part of the data to each mapper, and train individual models.
mus = zeros(num_mappers, size(train_data, 2));
Sigmas = cell(num_mappers, 1);
Sigmas_inv = cell(num_mappers, 1);
mapper_subsets = cvpartition(size(train_data, 1), 'kfold', num_mappers);
num_train_points = zeros(num_mappers, 1);
for mapper = 1 : num_mappers
    [mu, Sigma, Sigma_inv] = arow(train_data(mapper_subsets.test(mapper), :), train_labels(mapper_subsets.test(mapper)), [], [], r, iters);
    
    num_train_points(mapper) = sum(mapper_subsets.test(mapper));
    mus(mapper, :) = mu;
    Sigmas{mapper} = Sigma;
    Sigmas_inv{mapper} = Sigma_inv;
end;

%% Run the aggregation of mapper models on the reducer.
[mu_star, Sigma_star] = arow_reduce_aggregation(mus, Sigmas, Sigmas_inv, num_train_points / sum(num_train_points));

%% Finished training, test the model.
test_num_points = size(test_data, 1);
num_mistakes = 0;
preds = zeros(test_num_points, 1);
for i = 1 : test_num_points
    preds(i) = mu_star * test_data(i, :)';
    if (test_labels(i) * preds(i) <= 0)
        num_mistakes = num_mistakes + 1;
    end;
end;
accuracy = 100 * (1 - num_mistakes / test_num_points);
disp(['Accuracy: ', num2str(acc, 4), '%, correctly classified ', num2str(test_num_points - num_mistakes), ' of ', num2str(test_num_points), ' points.']);