function [mu, Sigma] = arow_reduce_aggregation(mus, Sigmas, Sigmas_inv, probs, mu_init, Sigma_init)

% function [mu, Sigma] = arow_reduce_aggregation(
%                      mus, Sigmas, Sigmas_inv, probs, mu_init, Sigma_init)
%
% Computation of aggregated Gaussian distribution from a number of provided
% Gaussians. Used in AROW-MapReduce in the reduce stage to aggregate
% different models computed on the mappers.
%
% Inputs: mus        - (M x D) matrix of per-mapper mean vectors, where M
%                      is number of mappers and D is vector dimensionality
%         Sigmas     - cell array of length M, where each element is a
%                      covariance matrix of size (D x D)
%         Sigmas_inv - cell array of length M, each element is an inverse
%                      of a covariance matrix of size (D x D); can be empty
%         probs      - vector of prior probabilities of per-mapper Gaussian
%                      distributions; needs to sum to 1
%         mu_init    - initial value of the output parameter mu
%         Sigma_init - initial value of the output parameter Sigma
%
% Output: mu         - D-dimensional mean vector of aggregated Gaussian
%         Sigma      - (D x D) covariance matrix of aggregated Gaussian
%
% Author: Nemanja Djuric (nemanja@temple.edu), according to the paper
%         Djuric N. et al., 'Distributed Confidence-Weighted Classification 
%         on MapReduce', IEEE BigData, Santa Clara, CA, USA, 2013.
% Date  : February 20th, 2016

REDUCER_ITERS = 5;

% If no inputs are provided, generate random 2-D Gaussians and find the
% distribution that minimizes the average weighted KL-divergence.
if (nargin == 0) || (isempty(mus))
    num_distributions = 3;
    mus = randn(num_distributions, 2);
    for i = 1 : 10
        A = randn(2);
        A = A * A';
        Sigmas{i} = A;
        Sigmas_inv{i} = inv(A);
    end;
    
    probs = rand(1, num_distributions);
    probs = probs / sum(probs);
%     probs = [0.99, 0.01, 0.01];
end;

[num_distributions, dim] = size(mus);

if (nargin < 5)
    mu_init = mean(mus);
    Sigma_init = eye(dim);
end;

mu = mu_init;
Sigma = Sigma_init;

% Find inverses of Sigmas if not already provided by the mappers.
if (isempty(Sigmas_inv))
    Sigmas_inv = cell(1, num_distributions);
    for i = 1 : num_distributions
        Sigmas_inv{i} = inv(Sigmas{i});
    end;
end;

% Solve the optimization problem on a reducer; equation (9) from the paper.
A = zeros(dim);
for i = 1 : num_distributions
    A = A + probs(i) * Sigmas_inv{i};
end;

for iter = 1 : REDUCER_ITERS
    Sigma_old = Sigma;
    
    B = zeros(dim);
    for i = 1 : num_distributions
        B = B + probs(i) * (Sigmas{i} + (mus(i, :) - mu)' * (mus(i, :) - mu));
    end;
    Sigma = solve_riccati(A, B);
    
    temp = zeros(dim);
    temp_mu = zeros(dim, 1);
    for i = 1 : num_distributions
        temp = temp + probs(i) * (Sigmas_inv{i} + inv(Sigma_old));
        temp_mu = temp_mu + probs(i) * (Sigmas_inv{i} + inv(Sigma_old)) * mus(i, :)';
    end;
    mu = (temp \ temp_mu)';
end;

% Plot the problem setting and solution (only if 2-dimensional problem,
% used just for illustration purposes).
if (dim == 2)
    figure, hold on
    for i = 1 : num_distributions
        h = plot_gaussian_ellipsoid(mus(i, :), Sigmas{i});
        set(h, 'LineWidth', 5 * probs(i));
    end;
    h = plot_gaussian_ellipsoid(mu, Sigma);
    set(h, 'color', 'r');
    set(h, 'LineWidth', 2);
    title({'Blue: Input Gaussians (thickness denotes'; 'prior probability); Red: Aggregated Gaussian'})
end;

function X = solve_riccati(A, B)
% Solves for X the Riccati equation XAX=B.
U = chol(A);
V = sqrtm(U*B*U');
R = inv(U);
X = R * V * R';